from unittest.mock import Mock

import pytest

from xent.benchmark.expand_benchmark import (
    expand_game_config,
    preprocess_dsl_code,
)
from xent.common.configuration_types import GameConfig
from xent.common.errors import XentSyntaxError
from xent.runtime.judge import Judge


@pytest.mark.integration
def test_generate_list_smoke():
    """Smoke test for JudgeGenerator.generate_list.

    Ensures the method returns a non-empty list of strings
    with at most the requested length and no obvious artifacts.
    """
    # Use a small, widely available model to keep the test light.
    try:
        judge = Judge("Qwen/Qwen3-14B-Base")
    except Exception as e:
        pytest.skip(f"Skipping generate_list smoke test (model unavailable): {e}")

    items = judge.text_generator.generate_list(
        "english nouns that have a prounoun of 'it' ie neuter objects", length=5
    )

    assert isinstance(items, list)
    assert len(items) == 5
    assert all(isinstance(x, str) and x.strip() for x in items)
    # No duplicates (case-insensitive) in the small sample
    assert len({x.strip().lower() for x in items}) == len(items)
    # Avoid leaking sentinel markers
    for x in items:
        lx = x.strip().lower()
        assert not lx.startswith("begin list")
        assert not lx.startswith("end list")


class TestExpandConfig:
    """Tests for expand_game_config and related functionality."""

    @pytest.fixture
    def mock_judge(self):
        """Create a mock Judge that returns predictable story content"""
        judge = Mock()
        # Use side_effect to return different values on subsequent calls
        judge.generate_text.side_effect = [
            "Once upon a time in a distant galaxy...",
            "The mysterious stranger arrived at midnight...",
            "In the depths of the ancient forest...",
            "A brilliant scientist made a discovery...",
            "The dragon soared above the clouds...",
        ]
        judge.set_seed = Mock()
        return judge

    @pytest.fixture
    def simple_game_config(self):
        """Create a simple game config with a single story() call"""
        return GameConfig(
            name="test_simple_story",
            code="""
                assign(s=story())
                reveal(s)
                elicit(x, 10)
                reward(xent(x | s))
            """,
            presentation_function="",
        )

    @pytest.fixture
    def multi_story_game_config(self):
        """Create a game config with multiple story() calls"""
        return GameConfig(
            name="test_multi_story",
            code="""
                assign(s1=story(), s2=story())
                reveal(s1, s2)
                elicit(x, 20)
                reward(xent(x | s1))
                assign(s3=story())
                reward(-xent(s3 | x))
            """,
            presentation_function="",
        )

    @pytest.fixture
    def no_story_game_config(self):
        """Create a game config without any story() calls"""
        return GameConfig(
            name="test_no_story",
            code="""
                assign(s="This is a hardcoded string")
                reveal(s)
                elicit(x, 15)
                reward(xent(x | s))
            """,
            presentation_function="",
        )

    @pytest.fixture
    def complex_story_game_config(self):
        """Create a game config with story() in complex expressions"""
        return GameConfig(
            name="test_complex_story",
            code="""
                assign(s1="Introduction: ")
                assign(s=s1 + story())
                reveal(s)
                assign(s2=(story() + " " + story()))
                elicit(response, 25)
                reward(xent(response | s2))
            """,
            presentation_function="",
        )

    def test_expand_game_config_simple_story(self, simple_game_config, mock_judge):
        """Test expansion of a simple game with one story() call"""
        map_seed = "test_seed_0"

        expanded = expand_game_config(simple_game_config, map_seed, mock_judge)

        # Verify structure
        assert isinstance(expanded, dict)
        assert expanded["name"] == "test_simple_story"
        assert expanded["map_seed"] == map_seed

        # Verify story() was replaced
        assert "story()" not in expanded["code"]
        assert "Once upon a time in a distant galaxy..." in expanded["code"]

        # Verify other code structure is preserved
        assert "assign(s=" in expanded["code"]
        assert "reveal(s)" in expanded["code"]
        assert "elicit(x, 10)" in expanded["code"]
        assert "reward(xent(x | s))" in expanded["code"]

        # Verify judge methods were called
        mock_judge.set_seed.assert_not_called()
        mock_judge.generate_text.assert_called_once()

    def test_expand_game_config_multiple_stories(
        self, multi_story_game_config, mock_judge
    ):
        """Test expansion of a game with multiple story() calls"""
        map_seed = "test_seed_1"

        expanded = expand_game_config(multi_story_game_config, map_seed, mock_judge)

        # Verify no story() calls remain
        assert "story()" not in expanded["code"]

        # Verify each story() was replaced with different content
        assert "Once upon a time in a distant galaxy..." in expanded["code"]
        assert "The mysterious stranger arrived at midnight..." in expanded["code"]
        assert "In the depths of the ancient forest..." in expanded["code"]

        # Should have generated 3 stories
        assert mock_judge.generate_text.call_count == 3

    def test_expand_game_config_no_story(self, no_story_game_config):
        """Test expansion of game with no story() calls"""
        mock_judge = Mock()
        expanded = expand_game_config(no_story_game_config, "seed", mock_judge)

        assert expanded["name"] == "test_no_story"
        assert "story()" not in expanded["code"]
        assert "This is a hardcoded string" in expanded["code"]

        # No generation should occur
        mock_judge.generate_text.assert_not_called()

    def test_complex_story_composition(self, complex_story_game_config):
        """Test that complex expressions with story() are handled correctly"""
        mock_judge = Mock()
        mock_judge.generate_text.return_value = "Generated story content"

        expanded = expand_game_config(complex_story_game_config, "seed", mock_judge)
        code = expanded["code"]

        # Should contain both stories and composition preserved
        assert code.count("Generated story content") >= 2
        assert "assign(s=s1 +" in code
        assert "assign(s2=" in code

    def test_preprocess_dsl_code(self):
        """Direct tests of preprocess_dsl_code behavior"""
        mock_judge = Mock()
        mock_judge.generate_text.return_value = "Test story"

        code = "assign(x=story())\nreveal(x)\nelicit(10)"
        result = preprocess_dsl_code(code, mock_judge)

        # All story() calls replaced and code structure preserved
        assert "story()" not in result
        assert "'Test story'" in result
        assert "reveal(x)" in result
        assert "elicit(10)" in result

        # Each line should be properly transformed
        lines = result.splitlines()
        assert lines[0] == "assign(x='Generated story content')" or (
            lines[0] == "assign(x='Test story')"
        )
        assert lines[1] == "reveal(x)"
        assert lines[2] == "elicit(10)"

    def test_story_rewriter_only_replaces_story_function(self):
        """Test that StoryRewriter only replaces story() calls, not other functions"""
        mock_judge = Mock()
        mock_judge.generate_text.return_value = "Test story"

        code_with_other_functions = """
            assign(a=story(), b=other_function(), c=yet_another())
            result = compute(story(), param)
        """

        result = preprocess_dsl_code(code_with_other_functions, mock_judge)

        # story() should be replaced
        assert "story()" not in result
        assert "'Test story'" in result

        # Other functions should remain unchanged
        assert "other_function()" in result
        assert "yet_another()" in result
        assert "compute(" in result

    def test_expand_game_config_empty_code(self):
        """Test expansion of game with empty code"""
        game_config = GameConfig(name="test_empty", code="", presentation_function="")

        mock_judge = Mock()
        map_seed = "test_seed"

        expanded = expand_game_config(game_config, map_seed, mock_judge)

        assert expanded["name"] == "test_empty"
        assert expanded["code"] == ""
        assert expanded["map_seed"] == map_seed

        # No story generation should occur
        mock_judge.generate_text.assert_not_called()

    def test_comments_and_formatting_comprehensive(self):
        """Comprehensive test for comment preservation and formatting in all scenarios."""
        # Test 1: Full-line comments
        judge1 = Mock()
        judge1.generate_text.return_value = "Once upon a time in a distant galaxy..."
        code_full_line = """
# This is a full-line comment.
assign(s=story())
# Another comment
        """
        processed_full = preprocess_dsl_code(code_full_line, judge1)
        assert "# This is a full-line comment." in processed_full
        assert "assign(s='Once upon a time in a distant galaxy...')" in processed_full
        assert "# Another comment" in processed_full

        # Test 2: Inline comments
        judge2 = Mock()
        judge2.generate_text.return_value = "Once upon a time in a distant galaxy..."
        code_inline = "assign(s=story())  # This is an inline comment."
        processed_inline = preprocess_dsl_code(code_inline, judge2)
        assert "assign(s='Once upon a time in a distant galaxy...')" in processed_inline
        assert "# This is an inline comment." in processed_inline

        # Test 3: Comment-only lines
        judge3 = Mock()
        code_comment_only = "# Just a comment"
        processed_comment_only = preprocess_dsl_code(code_comment_only, judge3)
        assert code_comment_only in processed_comment_only
        judge3.generate_text.assert_not_called()

        # Test 4: Empty lines preserved
        judge4 = Mock()
        judge4.generate_text.return_value = "Once upon a time in a distant galaxy..."
        code_empty_lines = """
assign(s=story())

reveal(s)
        """
        processed_empty = preprocess_dsl_code(code_empty_lines, judge4)
        assert "assign(s='Once upon a time in a distant galaxy...')" in processed_empty
        assert "\n\n" in processed_empty
        assert "reveal(s)" in processed_empty

        # Test 5: Mixed comments and code
        judge5 = Mock()
        judge5.generate_text.side_effect = [
            "Once upon a time in a distant galaxy...",
            "The mysterious stranger arrived at midnight...",
        ]
        code_mixed = """
# Header comment
assign(s1=story())  # First story
assign(s2=story())  # Second story

# Footer comment
        """
        processed_mixed = preprocess_dsl_code(code_mixed, judge5)
        assert "# Header comment" in processed_mixed
        assert "assign(s1='Once upon a time in a distant galaxy...')" in processed_mixed
        assert (
            "assign(s2='The mysterious stranger arrived at midnight...')"
            in processed_mixed
        )
        assert "# Footer comment" in processed_mixed

    def test_generate_list_rewrite_positional(self):
        """generate_list(prompt, length) is rewritten into a literal list."""
        mock_judge = Mock()

        def gl_side_effect(prompt: str, length: int) -> list[str]:
            assert prompt == "colors"
            assert length == 3
            return ["red", "blue", "green"]

        mock_judge.generate_list.side_effect = gl_side_effect

        code = 'assign(l=generate_list("colors", 3))'
        rewritten = preprocess_dsl_code(code, mock_judge)
        assert "generate_list(" not in rewritten
        assert "assign(l=[" in rewritten
        # Basic presence checks for returned items
        assert "red" in rewritten and "blue" in rewritten and "green" in rewritten

    def test_generate_list_rejects_keywords(self):
        """generate_list with keyword args should raise a syntax error."""
        mock_judge = Mock()
        code = 'assign(l=generate_list(length=2, prompt="animals"))'
        with pytest.raises(XentSyntaxError):
            preprocess_dsl_code(code, mock_judge)

    def test_generate_list_rejects_wrong_arity(self):
        mock_judge = Mock()
        code = 'assign(l=generate_list("colors"))'
        with pytest.raises(XentSyntaxError):
            preprocess_dsl_code(code, mock_judge)

        code2 = 'assign(l=generate_list("colors", 3, 4))'
        with pytest.raises(XentSyntaxError):
            preprocess_dsl_code(code2, mock_judge)

    def test_generate_list_rejects_wrong_types(self):
        mock_judge = Mock()
        code = "assign(l=generate_list(123, 3))"
        with pytest.raises(XentSyntaxError):
            preprocess_dsl_code(code, mock_judge)

        code2 = 'assign(l=generate_list("colors", "five"))'
        with pytest.raises(XentSyntaxError):
            preprocess_dsl_code(code2, mock_judge)

    def test_generate_list_accepts_float_length(self):
        """Numeric literal length (float) is accepted and cast to int."""
        mock_judge = Mock()

        def gl(prompt: str, length: int) -> list[str]:
            # Ensure length was cast to int
            assert isinstance(length, int) and length == 3
            assert prompt == "colors"
            return ["red", "blue", "green"]

        mock_judge.generate_list.side_effect = gl
        code = 'assign(l=generate_list("colors", 3.0))'
        rewritten = preprocess_dsl_code(code, mock_judge)
        assert "assign(l=[" in rewritten
        assert "red" in rewritten and "blue" in rewritten and "green" in rewritten

    def test_generate_list_rejects_nonliteral_prompt(self):
        mock_judge = Mock()
        code = 'assign(p="colors")\nassign(l=generate_list(p, 5))'
        with pytest.raises(XentSyntaxError):
            preprocess_dsl_code(code, mock_judge)

    def test_generate_list_rejects_swapped_types(self):
        mock_judge = Mock()
        code = 'assign(l=generate_list(5, "colors"))'
        with pytest.raises(XentSyntaxError):
            preprocess_dsl_code(code, mock_judge)



import json
import logging
import os
import subprocess

from xent.common.configuration_types import (
    ExecutableGameMap,
    ExpandedXentBenchmarkConfig,
)
from xent.common.token_xent_list import TokenXentList, ValidatedBool
from xent.common.x_list import XList
from xent.common.x_string import XString


class XEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, XString):
            return str(o)
        elif isinstance(o, XList):
            return list(o.items)
        elif isinstance(o, tuple):
            return {"__tuple__": True, "items": list(o)}
        elif isinstance(o, TokenXentList):
            return {"__TokenXentList__": True, "pairs": o.pairs, "scale": o.scale}
        elif isinstance(o, ValidatedBool):
            return bool(o)
        return super().default(o)


# No need to load XString from JSON
def x_decoder(dct):
    if "__tuple__" in dct:
        return tuple(dct["items"])
    elif "__TokenXentList__" in dct:
        return TokenXentList(dct["pairs"], dct["scale"])
    return dct


def dumps(obj, **kwargs):
    return json.dumps(obj, cls=XEncoder, **kwargs)


def loads(s, **kwargs):
    return json.loads(s, object_hook=x_decoder, **kwargs)


# Used to place git commit hash and diff in logs, which is useful for ensuring reproducibility of benchmark results
def log_git_snapshot(repo_path="."):
    if not os.path.isdir(repo_path):
        logging.warning(f"Git snapshot failed: Directory not found at '{repo_path}'")
        return

    try:
        commit_hash = subprocess.check_output(
            ["git", "rev-parse", "HEAD"],
            cwd=repo_path,
            stderr=subprocess.PIPE,
            text=True,
        ).strip()
        logging.info(f"Git Commit: {commit_hash}")

        git_diff = subprocess.check_output(
            ["git", "diff", "HEAD"], cwd=repo_path, stderr=subprocess.PIPE, text=True
        ).strip()

        if git_diff:
            logging.info(
                "Git Diff (staged/unstaged changes against HEAD):\n%s", git_diff
            )
        else:
            logging.info("Git Diff: No local changes detected against HEAD.")

        untracked_files = subprocess.check_output(
            ["git", "ls-files", "--others", "--exclude-standard"],
            cwd=repo_path,
            stderr=subprocess.PIPE,
            text=True,
        ).strip()

        if untracked_files:
            logging.info("Untracked files:\n%s", untracked_files)

    except (subprocess.CalledProcessError, FileNotFoundError) as e:
        logging.warning(
            "Could not retrieve Git snapshot. This might be because 'git' is not installed, "
            f"the path '{repo_path}' is not a Git repository, or it has no commits. Error: {e}"
        )
    except Exception as e:
        logging.error(f"An unexpected error occurred while logging git snapshot: {e}")


def generate_executable_game_maps(
    config: ExpandedXentBenchmarkConfig,
) -> list[ExecutableGameMap]:
    game_maps: list[ExecutableGameMap] = []
    for map in config["maps"]:
        for player in config["players"]:
            game_maps.append(
                ExecutableGameMap(
                    game_map=map, metadata=config["metadata"], player=player
                )
            )
    return game_maps



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from xent.common.errors import XentTypeError

if TYPE_CHECKING:
    from xent.common.x_list import XList


class XString:
    primary_string: str

    def __init__(
        self,
        primary_string: str | XString,
        static=False,
        public=False,
        name: str | None = None,
    ):
        if isinstance(primary_string, XString):
            primary_string = primary_string.primary_string

        if not isinstance(primary_string, str):
            raise XentTypeError(
                f"XString constructor requires a string argument. Got: {type(primary_string).__name__}"
            )

        self.primary_string = primary_string
        self.prefix = ""
        self.static = static
        self.public = public
        self.name = name

    def __str__(self):
        return str(self.primary_string)

    def __repr__(self):
        return f"XString('{self.primary_string}', prefix='{self.prefix}', static={self.static}, public={self.public}, name='{self.name}')"

    def serialize(self) -> dict[str, Any]:
        return {
            "type": "XString",
            "primary_string": self.primary_string,
            "prefix": self.prefix,
            "static": self.static,
            "public": self.public,
            "name": self.name,
        }

    @classmethod
    def deserialize(cls, data: dict[str, Any]):
        xstr = cls(data["primary_string"], data["static"], data["public"], data["name"])
        xstr.prefix = data["prefix"]
        return xstr

    def _verify_other_operand(self, other):
        if not isinstance(other, str) and not isinstance(other, XString):
            raise XentTypeError(
                f"Unsupported operand type(s): '{type(self).__name__}' and '{type(other).__name__}'. "
                "Operand must be a String."
            )

    def __eq__(self, other):
        if isinstance(other, XString):
            return self.primary_string == other.primary_string
        elif isinstance(other, str):
            return self.primary_string == other
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, XString):
            return self.primary_string != other.primary_string
        elif isinstance(other, str):
            return self.primary_string != other
        return NotImplemented

    def __or__(self, other):
        """This implements the prefix decoration operator |"""
        self._verify_other_operand(other)
        new_xstring = XString(self.primary_string)
        new_xstring.prefix = str(other)
        return new_xstring

    def __ror__(self, other):
        """This implements the prefix decoration operator |"""
        self._verify_other_operand(other)
        new_xstring = XString(str(other))
        new_xstring.prefix = str(self.primary_string)
        return new_xstring

    def __add__(self, other):
        self._verify_other_operand(other)
        return XString(self.primary_string + str(other))

    def __radd__(self, other):
        self._verify_other_operand(other)
        return XString(str(other) + self.primary_string)

    def _cut_front(self, left, right):
        if not right:
            return XString(left)

        try:
            index = left.index(right)
            return XString(left[:index])
        except ValueError:
            return XString(left)

    def __floordiv__(self, other):
        self._verify_other_operand(other)
        s_str = self.primary_string
        t_str = str(other)
        return self._cut_front(s_str, t_str)

    def __rfloordiv__(self, other):
        self._verify_other_operand(other)
        s_str = str(other)
        t_str = self.primary_string
        return self._cut_front(s_str, t_str)

    def _cut_back(self, left, right):
        if not right:
            return XString("")
        try:
            index = left.index(right)
            return XString(left[index + len(right) :])
        except ValueError:
            return XString("")

    def __mod__(self, other):
        """This implements the cut back operator %"""
        self._verify_other_operand(other)
        s_str = self.primary_string
        t_str = str(other)
        return self._cut_back(s_str, t_str)

    def __rmod__(self, other):
        """This implements the cut back operator %"""
        self._verify_other_operand(other)
        print("*****************IN RMOD*****************")
        print(f"Other: {other}")
        s_str = str(other)
        t_str = self.primary_string
        return self._cut_back(s_str, t_str)

    def __len__(self):
        return len(self.primary_string)

    def __contains__(self, item):
        return str(item) in self.primary_string

    def join(self, items: XList) -> XString:
        from xent.common.x_list import XList

        if not isinstance(items, XList):
            raise XentTypeError(
                f"XString.join requires an XList argument. Got: {type(items).__name__}"
            )

        joined_parts: list[str] = []
        for element in items:
            if not isinstance(element, XString):
                raise XentTypeError(
                    "XString.join requires all elements of the XList to be XString instances. "
                    f"Got: {type(element).__name__}"
                )
            joined_parts.append(element.primary_string)

        return XString(self.primary_string.join(joined_parts))

    def split(
        self,
        separator: str | XString | None = None,
    ) -> XList:
        from xent.common.x_list import XList

        if separator is None:
            sep_value: str | None = None
        elif isinstance(separator, XString):
            sep_value = separator.primary_string
        elif isinstance(separator, str):
            sep_value = separator
        else:
            raise XentTypeError(
                "XString.split separator must be None, str, or XString. "
                f"Got: {type(separator).__name__}"
            )

        split_result = self.primary_string.split(sep_value)

        return XList([XString(part) for part in split_result])



from abc import ABC, abstractmethod

from xent.common.configuration_types import (
    BenchmarkResult,
    ExpandedXentBenchmarkConfig,
    GameMapResults,
)


class Storage(ABC):
    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    async def list_configs(self) -> list[ExpandedXentBenchmarkConfig]:
        pass

    @abstractmethod
    async def add_config(self, config: ExpandedXentBenchmarkConfig):
        pass

    @abstractmethod
    async def list_result_ids(self) -> set[str]:
        pass

    @abstractmethod
    async def get_result(self, benchmark_id: str) -> BenchmarkResult | None:
        pass


class BenchmarkStorage(ABC):
    def __init__(self, benchmark_id: str):
        self.benchmark_id = benchmark_id

    @abstractmethod
    async def initialize(self):
        pass

    @abstractmethod
    async def clear(self):
        pass

    @abstractmethod
    async def get_config(self) -> ExpandedXentBenchmarkConfig | None:
        pass

    @abstractmethod
    async def store_config(self, config: ExpandedXentBenchmarkConfig):
        pass

    @abstractmethod
    async def get_game_map_results(
        self, game_name: str, map_seed: str, player_id: str
    ) -> GameMapResults | None:
        pass

    @abstractmethod
    async def store_game_map_results(self, results: GameMapResults):
        pass

    @abstractmethod
    async def get_benchmark_results(self) -> BenchmarkResult | None:
        pass

    @abstractmethod
    async def get_running_state(self) -> bool:
        pass

    @abstractmethod
    async def set_running_state(self, running: bool):
        pass



import logging
from collections.abc import Callable, Mapping
from typing import Any

from xent.common.configuration_types import XentEvent, XentMetadata
from xent.common.errors import XentConfigurationError, XentInternalError
from xent.common.version import get_xent_version
from xent.common.x_list import XList
from xent.common.x_string import XString
from xent.common.xent_event import LLMMessage

# A default example for presentation
DEFAULT_PRESENTATION = """
from typing import Any
from xent.presentation.sdk import (
    ChatBuilder,
    format_elicit_request,
    format_elicit_response,
    format_reveal,
    format_reward,
    format_failed_ensure,
)

def present_turn(state, since_events, metadata, full_history=None, ctx=None):
    if ctx is None:
        ctx = {}

    b = ChatBuilder()

    # Send a one-time introduction/instructions
    if not ctx.get('intro_sent', False):
        b.user(
            "You are playing a text game. Provide your move inside <move></move> tags. Any other text will be ignored."
        )
        ctx['intro_sent'] = True

    # Summarize what happened since the last elicit
    for event in since_events:
        t = event.get('type')
        if t == 'elicit_request':
            b.user(format_elicit_request(event))
        elif t == 'elicit_response':
            b.user(format_elicit_response(event))
        elif t == 'reveal':
            b.user(format_reveal(event))
        elif t == 'reward':
            formatted, _ = format_reward(event)
            b.user("Score update:\\n" + str(formatted))
        elif t == 'failed_ensure':
            b.user(format_failed_ensure(event))
        else:
            b.user("Event: " + str(event))

    # Close with a concise instruction cue
    b.user("Now provide your next move inside <move></move> tags.")

    return b.render(), ctx
"""


SAMPLE_METADATA: XentMetadata = XentMetadata(
    benchmark_id="bid",
    xent_version=get_xent_version(),
    judge_model="judge",
    num_rounds_per_game=2,
    seed="seed",
    store_full_player_interactions=False,
    npcs=[],
)


def get_default_presentation() -> str:
    return DEFAULT_PRESENTATION.strip()


class PresentationFunction:
    """
    Loader/executor for presentation functions that implement:

        present_turn(state, since_events, metadata, full_history=None, ctx=None)
            -> list[LLMMessage] | tuple[list[LLMMessage], dict[str, Any]]

    The __call__ normalizes the return into (messages, new_ctx).
    """

    def __init__(self, code_string: str):
        self.code_string = code_string
        self.compiled_code = None
        self.present_turn_func: (
            Callable[
                [
                    Mapping[str, Any],
                    list[XentEvent],
                    XentMetadata,
                    list[XentEvent] | None,
                    dict[str, Any] | None,
                ],
                Any,
            ]
            | None
        ) = None

        try:
            self.compiled_code = compile(code_string, "<presentation_turn>", "exec")
        except SyntaxError as e:
            raise XentInternalError(
                f"Turn presentation function syntax error: {e}"
            ) from e

        self.namespace: Any = {}

        try:
            exec(self.compiled_code, self.namespace)
        except Exception as e:
            raise XentConfigurationError(
                f"Error executing turn presentation function: {e}"
            ) from e

        if "present_turn" not in self.namespace:
            raise XentConfigurationError(
                "Turn presentation must define a 'present_turn' function"
            )

        present_turn_func = self.namespace["present_turn"]
        if not callable(present_turn_func):
            raise XentInternalError("'present_turn' must be a callable function")

        self.present_turn_func = present_turn_func  # type: ignore

    def __call__(
        self,
        state: Mapping[str, XString | XList],
        since_events: list[XentEvent],
        metadata: XentMetadata,
        full_history: list[XentEvent] | None = None,
        ctx: dict[str, Any] | None = None,
    ) -> tuple[list[LLMMessage], dict[str, Any]]:
        if self.present_turn_func is None:
            raise XentInternalError(
                "Turn presentation function not properly initialized"
            )

        try:
            result = self.present_turn_func(
                state, since_events, metadata, full_history, ctx
            )
            # Accept either list[LLMMessage] or (list[LLMMessage], ctx)
            messages: list[LLMMessage]
            new_ctx: dict[str, Any]

            if isinstance(result, tuple) and len(result) == 2:
                messages_candidate, ctx_candidate = result
                messages = (
                    list(messages_candidate) if messages_candidate is not None else []
                )
                new_ctx = (
                    dict(ctx_candidate)
                    if isinstance(ctx_candidate, dict)
                    else (ctx or {})
                )
            else:
                messages = list(result) if result is not None else []
                new_ctx = ctx or {}

            # Coerce message shapes to LLMMessage
            normalized: list[LLMMessage] = []
            for m in messages:
                role = getattr(m, "role", None) or (
                    m.get("role") if isinstance(m, dict) else None
                )
                content = getattr(m, "content", None) or (
                    m.get("content") if isinstance(m, dict) else None
                )
                if not role or not content:
                    logging.warning(f"Invalid message in turn presentation output: {m}")
                    continue
                normalized.append({"role": str(role), "content": str(content)})  # type: ignore[typeddict-item]

            return normalized, new_ctx
        except Exception as e:
            logging.error(f"Error in turn presentation function execution: {e}")
            raise XentConfigurationError(
                "Error in turn presentation function execution"
            ) from e

    def validate(
        self,
        sample_state: dict[str, Any] | None = None,
        sample_since_events: list[XentEvent] | None = None,
        sample_metadata: XentMetadata | None = None,
    ) -> bool:
        if sample_state is None:
            sample_state = {}
        if sample_since_events is None:
            sample_since_events = []
        if sample_metadata is None:
            sample_metadata = SAMPLE_METADATA

        try:
            messages, _ctx = self(sample_state, sample_since_events, sample_metadata)
            valid = isinstance(messages, list)
            return valid
        except Exception as e:
            logging.error(f"Presentation function validation failed: {e}")
            return False



from importlib.metadata import PackageNotFoundError, version


def get_xent_version() -> str:
    try:
        return version("xent")
    except PackageNotFoundError:
        # Package is not installed (likely in development/editable mode)
        return "0.3.0-dev"


def validate_version(
    config_version: str | None, current_version: str
) -> tuple[bool, str]:
    if config_version is None:
        return (
            True,
            "Warning: Configuration has no version field (likely created with older xent version)",
        )

    if config_version == current_version:
        return True, f"Version match: {current_version}"

    return (
        False,
        f"Version mismatch: configuration was created with xent {config_version}, "
        f"but current version is {current_version}",
    )
